@@ -1,3 +1,4 @@
+import fortpy
 from .generator import TestGenerator
 from ..code import CodeParser, secondsToStr
 from os import system, path, mkdir
@@ -11,7 +12,7 @@ class ExecutionResult(object):
     """The result of running the executable on the system, NOT the
     result of the unit test file comparisons.
 
-    :arg execfolder: the folder in which the executable ran.
+    :arg folder: the folder in which the executable ran.
     :arg exitcode: the system exit code after the process terminated.
     :arg runtime: a python datetime for the execution run time.
     :arg tester: an OutcomeTester that can test the outcome of this
@@ -29,17 +30,17 @@ class ExecutionResult(object):
     def _compare_time(self):
         """Determines if the runtime for this execution exceeded the
         maximum in the tester."""
-        if self.tester.runtime is not None and \
-           self.tester.unit is not None:
+        if self.tester.testspec.runtime is not None and \
+           self.tester.testspec.unit is not None:
             #Get the runtime in a nice-to-compare format 0:00:00.000
             stime = secondsToStr(self.runtime)
             #we have limits specified as minutes or hours in a range
             #No unit test should/can take more than 24 hours.
             utime = [ int(u) for u in stime.split(":") ]
-            if self.tester.unit == "h":
-                return self._compare_list_time(self.tester.runtime, utime[0])
-            elif self.tester.unit == "m":
-                return self._compare_list_time(self.tester.runtime, utime[1])  
+            if self.tester.testspec.unit == "h":
+                return self._compare_list_time(self.tester.testspec.runtime, utime[0])
+            elif self.tester.testspec.unit == "m":
+                return self._compare_list_time(self.tester.testspec.runtime, utime[1])  
         else:
             return True
 
@@ -54,7 +55,7 @@ class ExecutionResult(object):
         self.tester.test(caseid, self, uresult)
         #All that's left is to check the runtime against its max.
         if not self._compare_time():
-            uresult.overtimes[caseid] = (self.runtime, self.tester.runtime)        
+            uresult.overtimes[caseid] = (self.runtime, self.tester.testspec.runtime)
 
 class ValueCompareResult(object):
     """The result of comparing the contents of a file to an explicit
@@ -83,86 +84,45 @@ class ValueCompareResult(object):
         return result
 
 class OutcomeTester(object):
-    """Performs outcomes tests for a single 'outcome' DocString.
+    """Performs outcomes tests for a single 'test' DocString.
 
-    :arg doc: the DocString for the 'outcome' tag.
+    :arg testspec: the TestSpecification for the <test> tag.
     :arg codefolder: the path to the folder that has all the modules from which
       the unit tests were built.
     :arg comparer: an instance of FileComparer to compare output files
       to the model ones.
     """
-    def __init__(self, doc, codefolder, comparer, verbose):
-        self.doc = doc
+    def __init__(self, testspec, codefolder, comparer, verbose):
+        self.testspec = testspec
+        self.codefolder = codefolder
         self.comparer = comparer
         self.verbose = verbose
-
-        #Extract the paths to the source input and output folders
-        self.source = doc.attributes["sourcepath"].replace(".", codefolder)
-        if "comparepath" in doc.attributes:
-            self.compare = doc.attributes["comparepath"].replace(".", codefolder)
-        else:
-            self.compare = self.source
-        
-        #Extract the targets that need to be compared and the files to compare
-        #them to. Since multiple targets can be specified, we need to look at 
-        #each one to see if it is a file or a variable target
-
-        #The corresponding entries in the compare attribute must be either files
-        #or values too.
-        if "target" not in doc.attributes or "compare" not in doc.attributes:
-            print("WARNING: either targets or model output files were missing for {}".format(doc))
-            self.targets = []
-            self.models = []
-        else:
-            self.targets = doc.attributes["target"].split(",")
-            self.models = doc.attributes["compare"].split("?")
-
-        #Get all the other optional attributes
-        self._parse_xml()
-
-    def _parse_xml(self):
-        """Extracts XML attributes related to comparison modes and tolerances
-        that are used to analyze the file comparison."""
-        #Extract the doc attributes related to the comparison performance
-        if "mode" in self.doc.attributes:
-            self.mode = self.doc.attributes["mode"]
-        else:
-            self.mode = "default"
-
-        if "runtime" in self.doc.attributes:
-            runtime = list(self.doc.attributes["runtime"])
-            self.unit = runtime.pop()
-            self.runtime = [ int(t) for t in "".join(runtime).split("-") ]
-        else:
-            self.runtime = None
-            self.unit = None
-
-        if "tolerance" in self.doc.attributes:
-            self.tolerance = float(self.doc.attributes["tolerance"])
-        else:
-            self.tolerance = 1
         
     def test(self, caseid, xresult, uresult):
-        """Checks the output of the execution for the specified outcome doctag.
+        """Checks the output of the execution for the specified test specification.
 
         :arg xresult: an ExecutionResult instance that has information about the
           execution whose outcome is being tested.
         :arg uresult: the unittest's TestResult that holds information from all
           executions that were run for different cases.
         """        
-        #Now compare the files with their model outputs
+        if not self.testspec.testable:
+            return
+
+        #Compare the files with their model outputs
         compresults = []
-        for i in range(len(self.models)):
-            paths = self._run_get_paths(self.targets[i], self.models[i], xresult.folder,
-                                        self.compare, xresult.case)
+        for i in range(len(self.testspec.targets)):
+            target = self.testspec.targets[i]
+            exepath, outvar, isfile = self._run_get_paths(target, xresult.folder)
+
             #The third entry in the tuple identifies it as an explicit value
             #comparison as opposed to a file comparison.
-            if paths[2]:
-                result = self._run_compare_file(paths[0], paths[1])
-                if result is None or result.common_match < self.tolerance:
+            if isfile:
+                result = self._run_compare_file(exepath, outvar, self.codefolder, caseid)
+                if result is None or result.common_match < outvar.tolerance:
                     self._add_failure(caseid, result, uresult)
             else:
-                result = self._run_compare_var(paths[0], eval(paths[1]))
+                result = self._run_compare_var(exepath, outvar)
                 if result is None or not result.equal:
                     self._add_failure(caseid, result, uresult)
             
@@ -178,25 +138,33 @@ class OutcomeTester(object):
         else:
             uresult.failures[caseid] = [ result ]
 
-    def _run_compare_var(self, exe, value, doc):
+    def _run_compare_var(self, exepath, outvar):
         """Compares the value of a variable written to an output file with an
-        explicit value."""
-        return ValueCompareResult(exe, value)
+        explicit value.
 
-    def _run_compare_file(self, exe, model):
+        :arg exepath: the full path to the file created by the unit test executable.
+        :arg outvar: the TestOutput instance with testing specifications.
+        """
+        return ValueCompareResult(exe, outvar.value)
+
+    def _run_compare_file(self, exepath, outvar, coderoot, caseid):
         """Compares an output file from an executable with its model output using
         settings in the doctag.
 
         :arg exe: the full path to the output file from the unit test.
-        :arg model: the full path to the model output file to compare.
+        :arg outvar: the TestOutput instance with testing specifications.
+        :arg coderoot: the full path to the folder that has all the code files.
+        :arg caseid: the identifier for the specific test case being tested.
         """
         #First get the comparison results, then analyze them using the tolerances
         #etc. from the doctag.
-        result = self.comparer.compare(exe, model, self.mode)
+        targetpath = outvar.abspath(coderoot, caseid)
+        result = self.comparer.compare(exepath, targetpath, outvar.template, outvar.mode)
+
         #Write the results out to file as a record. If the result is none create
         #a file that says so. The default file name is the output file name with
         #an extra extension of .compare
-        resultpath = exe + ".compare"
+        resultpath = exepath + ".compare"
         with open(resultpath, "w") as f:  
             if result is not None:          
                 f.write(print_compare_result(result, self.verbose))
@@ -205,41 +173,38 @@ class OutcomeTester(object):
             
         return result
 
-    def _run_get_paths(self, target, compare, exefolder, source, case):
+    def _run_get_paths(self, target, exefolder):
         """Gets the file paths to the executable output and model output
         files that need to be compared. If the model output is not a file
         the third entry in the tuple will be False.
 
         :arg target: the target file or variable that needs to be compared.
-        :arg compare: the name of the file in the source folder to which
-          the target will be compared.
         :arg execfolder: the folder in which the executable ran, also where
           the output files from the executable reside.
-        :arg source: the path to the source folder where the model outputs are.
+        :arg caseid: the identifier for the specific test case that ran.
         """
-        #Keep track of whether we are comparing a variable value file to a hard-coded
-        #value or to another output file.
-        isvar = False
-
-        if target[0] == ".":
-            exe = path.join(exefolder, target[2::]).format(case)
+        if target.name[0] == ".":
+            exe = path.join(exefolder, target.name[2::])
         else:
             #This is a variable, get the auto-generated filename
-            filename = target.replace("%", ".")
-            exe = path.join(exefolder, filename)
+            exe = path.join(exefolder, target.varfile)
 
-        if compare[0] == ".":
-            mod = path.join(source, compare[2::].format(case))
+        #In order to run the comparison, we must have an <output> tag to compare
+        #the target to.
+        if target.compareto in self.testspec.outputs:
+            outvar = self.testspec.outputs[target.compareto]
         else:
-            isvar = True
-            mod = compare
+            raise ValueError("Target's compareto='{}' ".format(target.compareto) + 
+                             "does not match any <output> tag.")
 
-        return (exe, mod, not isvar)     
+        return (exe, outvar, outvar.filemode)
 
 class TestResult(object):
     """Represents a set of unit test results.
 
     :arg identifier: the module.method identifier for this unit test.
+    :arg testid: the identifier of the <test> tag for the specific test that this
+      result represents.
     :attr cases: a dictionary of ExecutionResult objects with detail
       on how the system execution went for each case. Key is caseId
     :attr outcomes: a dictionary of CompareResult objects with detail
@@ -247,12 +212,19 @@ class TestResult(object):
       are the same caseIds used in self.cases.
     :attr compiled: specifies whether the executable associated with
       this identifier compiled successfully.
+    :attr failures: CompareResult or ValueResult instances that failed the
+      tests outlined by the relevant test specification. Keys are the
+      caseIds used by all the other dicts.
+    :attr overtimes: a tuple of (actual runtime, desired runtime range) for
+      tests that were outside of the range for runtime.
     """
-    def __init__(self, identifier):
+    def __init__(self, identifier, testid):
         self.identifier = identifier
+        self.testid = testid
         self.cases = {}
         self.outcomes = {}
         self.compiled = None
+
         self.warnings = []
         self.failures = {}
         self.overtimes = {}
@@ -293,13 +265,15 @@ class UnitTester(object):
     :arg libraryroot: the path to folder in which to stage the tests.
     """
     def __init__(self, libraryroot, verbose = False, compare_templates = None, 
-                 fortpy_templates = "~/pythonpkg/fortpy/templates", rerun = False):
+                 fortpy_templates=None, rerun = False, compiler="gfortran"):
         self.libraryroot = path.abspath(libraryroot)
         self.parser = CodeParser()
         self.parser.verbose = verbose
-        self.fortpy_templates = path.abspath(fortpy_templates)
-        self.tgenerator = TestGenerator(self.parser, self.libraryroot, self.fortpy_templates,
-                                        rerun)
+        self._set_fortpy_dir(fortpy_templates)
+        self.tgenerator = TestGenerator(self.parser, self.libraryroot, 
+                                        self.fortpy_templates, rerun)
+        self.compiler = compiler
+        self._compiler_exists()
         
         #A flag to track whether the generator has already written
         #the executables.
@@ -307,6 +281,61 @@ class UnitTester(object):
         #The user's raw value for the compare_templates directory
         self._compare_templates = compare_templates
 
+    @property
+    def tests(self):
+        """Returns a dictionary of all the tests that need to be run for the
+        current configuration of the test generator."""
+        return self.tgenerator.xgenerator.writer.tests
+
+    @property
+    def writer(self):
+        """Returns the underlying executable writer that has a list of ordered
+        method/assignment dependencies."""
+        return self.tgenerator.xgenerator.writer
+        
+    def _compiler_exists(self):
+        """Tests whether the specified compiler is available on the machine. If
+        it isn't give an error and exit."""
+        if self.which(self.compiler) is None:
+            print("ERROR: compiler {} not found. Exiting.".format(self.compiler))
+            exit(1)
+        
+    def which(self, program):
+        """Tests whether the specified program is anywhere in the environment
+        PATH so that it probably exists."""
+        import os
+        def is_exe(fpath):
+            return os.path.isfile(fpath) and os.access(fpath, os.X_OK)
+
+        fpath, fname = os.path.split(program)
+        if fpath:
+            if is_exe(program):
+                return program
+        else:
+            for path in os.environ["PATH"].split(os.pathsep):
+                path = path.strip('"')
+                exe_file = os.path.join(path, program)
+                if is_exe(exe_file):
+                    return exe_file
+
+        return None
+
+    def get_fortpy_templates_dir(self):
+        """Gets the templates directory from the fortpy package."""
+        fortdir = path.dirname(fortpy.__file__)
+        return path.join(fortdir, "templates")
+
+    def _set_fortpy_dir(self, fortpy_templates=None):
+        """Sets the directory path for the fortpy templates. If no directory
+        is specified, use the default one that shipped with the package.
+        """
+        #If they didn't specify a custom templates directory, use the default
+        #one that shipped with the package.
+        if fortpy_templates is not None:
+            self.fortpy_templates = path.abspath(fortpy_templates)
+        else:
+            self.fortpy_templates = self.get_fortpy_templates_dir()
+
     def writeall(self, codefolder):
         """Writes all the unit test executables that are new or modified
         as well as the makefiles for all subroutines in all modules."""
@@ -330,7 +359,7 @@ class UnitTester(object):
 
         #Now that we have loaded all the codefiles in the path, we can
         #generate the unittest executables
-        self.tgenerator.write()
+        self.tgenerator.write(self._codefolder)
         self._written = True
 
     def runall(self):
@@ -338,46 +367,47 @@ class UnitTester(object):
         After the run is complete, the outcomes are checked for consistency."""
         if self._written:
             #We will likely need a file comparer for the outcomes testing
-            self.comparer = FileComparer(self.compare_templates)
+            self.comparer = FileComparer(self.fortpy_templates, self.compare_templates)
             #Run them each individually and return a dictionary of all the
             #test results
             result = {}
-            for identifier in self.tgenerator.tests_to_run:
-                result[identifier] = self._run_single(identifier)
+            for composite_id in self.tgenerator.tests_to_run:
+                identifier, testid = composite_id.split("|")
+                result[composite_id] = self._run_single(identifier, testid)
 
             return result
         else:
             print("WARNING: you can't run tests until the executables have been written. Exiting.")
             return None
        
-    def _run_single(self, identifier):
+    def _run_single(self, identifier, testid):
         """Runs all unit test cases for the specified identifier."""
         #Just initialize a result object and populate its properties
         #using the various _run_* methods.
-        result = TestResult(identifier)
-        result.compiled = self._run_compile(identifier)
+        result = TestResult(identifier, testid)
+        result.compiled = self._run_compile(identifier, testid)
         if result.compiled:
-            self._run_exec(identifier, result)
+            self._run_exec(identifier, testid, result)
 
         return result
 
-    def _run_compile(self, identifier):
+    def _run_compile(self, identifier, testid):
         """Compiles the executable that was created for the specified identifier,
         returns True if the compile was successful."""
         #Find the target folder that has the executables etc then run
         #make and check the exit code.
         target = path.join(self.libraryroot, identifier)
         print("\n\n")
-        code = system("cd {}; make F90=ifort".format(target))
+        code = system("cd {}; make -f 'Makefile.{}' F90={}".format(target, testid, self.compiler))
         print("\n")
         return code == 0
 
-    def _run_exec(self, identifier, result):
+    def _run_exec(self, identifier, testid, result):
         """Runs the executable for unit test for the specified identifier
         for each of the outcomes specified in the doctags."""
         #Get the home path of the executable. A sub-folder for tests
         #needs to be created. For tests that have input and output files
-        #a home/tests/case folder gets created and the source files
+        #a home/tests/testid.case folder gets created and the source files
         #get copied.
 
         #Create the folder for staging the tests.
@@ -391,12 +421,10 @@ class UnitTester(object):
         method = module.executables[kmethod]
 
         #Get the absolute path to the executable that we created
-        exepath = path.join(self.libraryroot, identifier, "run.x")
-
-        for doc in method.tests:
-            if doc.doctype == "outcome"  and "target" in doc.attributes and \
-               "sourcepath" in doc.attributes:
-                self._run_folder(doc, tests, result, exepath)
+        exepath = path.join(self.libraryroot, identifier, "{}.x".format(testid))
+        #Since we have already split out all the tests that need to be run and 
+        #we have a 'testid' for the current test to run, just run that test.
+        self._run_folder(self.tests[testid], tests, result, exepath)
 
         #Now that we have run all of the executables, we can analyze their
         #output to see if it matches.
@@ -404,70 +432,71 @@ class UnitTester(object):
             xres = result.cases[case]
             xres.test(case, result)
 
-    def _run_folder(self, doc, testsfolder, result, exepath):
+    def _run_folder(self, testspec, testsfolder, result, exepath):
         """Runs the executable for the sources in the folder doctag.
 
-        :arg doc: the docstring 'outcome' that motivates the execution.
+        :arg testspec: a TestSpecification instance for this unit test.
         :arg testsfolder: the path to the unit tests unique staging folder.
         :arg result: a TestResult instance that execution results can be
           added to as they complete.
+        :arg expath: the full path to the executable to run in the folder.
         """
-        #The target path from the tag is applicable for all test types.
-        source = doc.attributes["sourcepath"].replace(".", self._codefolder)
-        #Renames is a linked list of target names for the files in sources.
-        if "rename" in doc.attributes:
-            renames = doc.attributes["rename"].split(",")
-        else:
-            renames = []
-
-        system("cd {}".format(source))
-
         #We can use the same outcome tester afterwards for all of the cases.
-        tester = OutcomeTester(doc, self._codefolder, self.comparer, self.parser.verbose)
+        tester = OutcomeTester(testspec, self._codefolder, self.comparer, self.parser.verbose)
 
         #The execution can either be case-based or once-off.
-        if "cases" in doc.attributes and "sources" in doc.attributes:
+        if testspec.cases is not None and len(testspec.inputs) > 0:
             #We need to run the executable multiple times, once for each case
             #Each case has input files specified relative to the code folder.
-            cases = doc.attributes["cases"].split(",")
-            for case in cases:
-                caseid = "/".join([doc.attributes["sourcepath"].replace(".",""), case])
+            for case in testspec.cases:
+                caseid = "{}.{}".format(testspec.identifier, case)
                 if not caseid in result.cases:
-                    inputs = doc.attributes["sources"].split(",")
-                    #Get the paths to the input files needed to run each case
-                    copies = [ path.join(source, i.format(case)) for i in inputs ]
-
                     #Make a separate directory for the case and copy all its inputs.
-                    casepath = path.join(testsfolder, case)
+                    casepath = path.join(testsfolder, caseid)
                     if not path.exists(casepath):
                         mkdir(casepath)
 
-                    #Copy all the input files we need to run this case. If a rename
-                    #attribute was specified, change the name of the target files.
-                    for i in range(len(copies)):
-                        f = copies[i]
-                        if len(renames) > 0 and i < len(renames):
-                            new = path.join(casepath, renames[i])
-                            copyfile(f, new)
-                        else:
-                            copy(f, casepath)
+                    for i in testspec.inputs:
+                        i.copy(self._codefolder, casepath, case)
+                    #Also copy any assignment file dependencies.
+                    self.writer.copy(self._codefolder, casepath, case)
+                    #Clean the testing folder to remove any target variable output files
+                    #from any earlier test runs.
+                    testspec.clean(casepath)
 
-                    print("Executing run.x for case {}".format(caseid))
+                    print("Executing {}.x for case {}".format(testspec.identifier, case))
 
                     #We need to time the execution as one of the possible outcome tests.
                     start_time = clock()                    
                     code = system("cd {}; {}".format(casepath, exepath))
-                    result.cases[caseid] = ExecutionResult(casepath, code, 
-                                                           clock() - start_time, tester, case)
+                    result.cases[caseid] = ExecutionResult(casepath, code, clock() - start_time, 
+                                                           tester, case)
+
                     #Add some whitespace for readability between tests
                     print("")
                 else:
                     result.warnings.append("Duplicate CASES specified for unit testing: {}".format(caseid))
         else:
-            print("Executing run.x in explicit folder {}".format(target))
-            start_time = clock()                    
-            code = system("cd {}; {}".format(testsfolder, exepath))
-            result.cases[doc.attributes["sourcepath"].replace(".","")] = ExecutionResult(
-                testsfolder, code, clock() - start_time, tester)
+            #Create a folder for this test specification to run in.
+            testpath = path.join(testsfolder, testspec.identifier)
+            if not path.exists(testpath):
+                mkdir(testpath)
+            print("Executing {}.x in explicit folder {}".format(testspec.identifier, testpath))
+
+            #Copy across all the input files we need to run.
+            for i in testspec.inputs:
+                i.copy(self._codefolder, testpath)
+            #Also copy any assignment file dependencies.
+            self.writer.copy(self._codefolder, testpath, "")
+            #Clean the testing folder to remove any target variable output files
+            #from any earlier test runs.
+            testspec.clean(testpath)
+
+
+            start_time = clock()                              
+            code = system("cd {}; {}".format(testpath, exepath))
+            result.cases[testspec.identifier] = ExecutionResult(
+                testpath, code, clock() - start_time, tester)
+
             #Add some whitespace for readability between tests
             print("")        
